@@ -21,7 +21,7 @@ def getData(greenbyte_client,QUERY_PARAMS,LOGGER):
         try:
             if tries==0:
                 LOGGER.debug(f"Running getData with parameters {QUERY_PARAMS}")
-            return greenbyte_client.data.get_data(QUERY_PARAMS['deviceIds'],QUERY_PARAMS['signalIds'],QUERY_PARAMS['timestampStart'],QUERY_PARAMS['timestampFinish'],False,QUERY_PARAMS['resolution'],'device',0,None)
+            return greenbyte_client.data.get_data(QUERY_PARAMS['deviceIds'],QUERY_PARAMS['dataSignalIds'],QUERY_PARAMS['timestampStart'],QUERY_PARAMS['timestampEnd'],False,QUERY_PARAMS['resolution'],'device',0,None)
         except Exception as ge:
             tries+=1
             LOGGER.debug(f"WARNING: get_data from Breeze not received, try {tries}, devices: {QUERY_PARAMS['deviceIds']}, error: "+str(ge))
@@ -37,22 +37,22 @@ def getStatuses(greenbyte_client,QUERY_PARAMS,LOGGER):
     while tries<3:
         try:
             if tries==0:
-                LOGGER.debug(f"Running getStatuses with parameters {QUERY_PARAMS}")
+                LOGGER.debug(f"Running gbtGlobal.getStatuses with parameters {QUERY_PARAMS}")
             response = greenbyte_client.statuses.get_statuses(QUERY_PARAMS['deviceIds'],QUERY_PARAMS['timestampStart'],QUERY_PARAMS['timestampFinish'],QUERY_PARAMS['categories'],
                                                           QUERY_PARAMS['lostProductionSignalId'],None,None,False,QUERY_PARAMS['pageSize'],QUERY_PARAMS['page'],False,'custom')
             if type(response) is list:
                 return response
             else:
-                LOGGER.error(f"getStatuses returned a non-list response. Check parameters of the greenbyte_client.")
+                LOGGER.error(f"ERROR: gbtGlobal.getStatuses from Breeze returned a non-list response. Check parameters of the greenbyte_client.")
                 return None
         except Exception as ge:
             tries+=1
             LOGGER.debug(f"WARNING: get_statuses from Breeze not received, try {tries}, devices: {QUERY_PARAMS['deviceIds']}, error: "+str(ge))
             if tries==3:
                 if  "Read timed out" in str(ge):
-                    LOGGER.error(f"ERROR: get_statuses from Breeze timed out for devices: {QUERY_PARAMS['deviceIds']}. Full error: "+str(ge))
+                    LOGGER.error(f"ERROR: gbtGlobal.getStatuses from Breeze timed out for devices: {QUERY_PARAMS['deviceIds']}. Full error: "+str(ge))
                 else:
-                    LOGGER.error(f"ERROR: get_statuses from Breeze not received, devices: {QUERY_PARAMS['deviceIds']}, error: "+str(ge))
+                    LOGGER.error(f"ERROR: gbtGlobal.getStatuses from Breeze not received, devices: {QUERY_PARAMS['deviceIds']}, error: "+str(ge))
                 return None
             
 def getSiteAccesses(greenbyte_client,QUERY_PARAMS,LOGGER):
