@@ -2,7 +2,7 @@ import logging
 from .TableLogHandler import TableLogHandler
 
 class Log:
-    def __init__(self, CREDS, log_table, log_level):
+    def __init__(self, CREDS, log_table, log_level, script=None):
         #logger.setLevel(logging.DEBUG)
         #notset=0
         #trace=5\does not exist!
@@ -15,8 +15,13 @@ class Log:
             CREDS["instrumentation_key"]
         except:
             CREDS["instrumentation_key"]=None
+
+        if script is None:
+            logger_var = log_table
+        else:
+            logger_var = script
         
-        self.logger = logging.getLogger(log_table)
+        self.logger = logging.getLogger(logger_var)
         self.logger.setLevel(int(log_level))
         if len(self.logger.handlers)>0:
             self.logger.handlers.clear()
